@@ -16,19 +16,23 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <fcntl.h>
 #include <assert.h>
 #include <math.h>
 
+#define USE_FC_LEN_T
+#include <Rconfig.h>
+#include "R_ext/BLAS.h"
+#ifndef FCONE
+# define FCONE
+#endif
 #include <R.h>
 #define USE_RINTERNALS
 #include <Rinternals.h>
 #include <Rdefines.h>
 
-#include "R_ext/BLAS.h"
 #include "R_ext/Lapack.h"
 #include "R_ext/Rdynload.h"
 #include "R_ext/Utils.h"
@@ -56,7 +60,10 @@ RNORM (int n)
       error ("invalid call");
     }
   for (int i = 0; i < length (cmdexpr); i++)
-    ans = eval (VECTOR_ELT (cmdexpr, i), R_GlobalEnv);
+    {
+      ans = PROTECT (eval (VECTOR_ELT (cmdexpr, i), R_GlobalEnv));
+      UNPROTECT (1);
+    }
   UNPROTECT (2);
   return ans;
 }
@@ -69,7 +76,7 @@ RNORM (int n)
  * WORK integer working subspace dimension must be > NU
  * MAXIT integer maximum number of iterations
  * TOL double tolerance
- * EPS double machine epsilon
+ * EPS double invariant subspace detection tolerance
  * MULT integer 0 X is a dense matrix (dgemm), 1 sparse (cholmod)
  * RESTART integer 0 no or > 0 indicates restart of dimension n
  * RV, RW, RS optional restart V W and S values of dimension RESTART
@@ -212,7 +219,7 @@ irlb (double *A,                // Input data matrix (double case)
       double *V,                // output right singular vectors n x work
       int *ITER,                // ouput number of Lanczos iterations
       int *MPROD,               // output number of matrix vector products
-      double eps,               // machine epsilon
+      double eps,               // tolerance for invariant subspace detection
       // working intermediate storage, sizes shown
       int lwork, double *V1,    // n x work
       double *U1,               // m x work
@@ -255,7 +262,7 @@ irlb (double *A,                // Input data matrix (double case)
       if (iter == 0 && restart == 0)
         {
           d = F77_NAME (dnrm2) (&n, V, &inc);
-          if (d < 2 * eps)
+          if (d < eps)
             return -1;
           d = 1 / d;
           F77_NAME (dscal) (&n, &d, V, &inc);
@@ -282,11 +289,10 @@ irlb (double *A,                // Input data matrix (double case)
           alpha = 1;
           beta = 0;
           F77_NAME (dgemv) ("n", &m, &n, &alpha, (double *) A, &m, x,
-                            &inc, &beta, W + j * m, &inc);
+                            &inc, &beta, W + j * m, &inc FCONE);
         }
       mprod++;
       R_CheckUserInterrupt ();
-
 /* optionally apply shift in square cases m = n */
       if (shift)
         {
@@ -302,11 +308,10 @@ irlb (double *A,                // Input data matrix (double case)
           for (kk = 0; kk < m; ++kk)
             W[jj + kk] = W[jj + kk] - beta;
         }
-
       if (iter > 0)
         orthog (W, W + j * m, T, m, j, 1);
       S = F77_NAME (dnrm2) (&m, W + j * m, &inc);
-      if (S < 2 * eps && j == 0)
+      if (S < eps && j == 0)
         return -4;
       SS = 1.0 / S;
       F77_NAME (dscal) (&m, &SS, W + j * m, &inc);
@@ -323,7 +328,7 @@ irlb (double *A,                // Input data matrix (double case)
               alpha = 1.0;
               beta = 0.0;
               F77_NAME (dgemv) ("t", &m, &n, &alpha, (double *) A, &m,
-                                W + j * m, &inc, &beta, F, &inc);
+                                W + j * m, &inc, &beta, F, &inc FCONE);
             }
           mprod++;
           R_CheckUserInterrupt ();
@@ -343,8 +348,12 @@ irlb (double *A,                // Input data matrix (double case)
             {
               beta = 0;
               for (kk = 0; kk < m; ++kk) beta += W[j *m + kk];
-              for (kk = 0; kk < n; ++kk)
-                F[kk] = F[kk] - beta * center[kk]; // XXX XXX XXX
+              if (scale)
+                for (kk = 0; kk < n; ++kk)
+                  F[kk] = F[kk] - beta * center[kk] / scale[kk];
+              else
+                for (kk = 0; kk < n; ++kk)
+                  F[kk] = F[kk] - beta * center[kk];
             }
           SS = -S;
           F77_NAME (daxpy) (&n, &SS, V + j * n, &inc, F, &inc);
@@ -354,7 +363,7 @@ irlb (double *A,                // Input data matrix (double case)
             {
               R_F = F77_NAME (dnrm2) (&n, F, &inc);
               R = 1.0 / R_F;
-              if (R_F < 2 * eps)        // near invariant subspace
+              if (R_F < eps)        // near invariant subspace
                 {
                   FOO = RNORM (n);
                   for (kk = 0; kk < n; ++kk)
@@ -364,7 +373,6 @@ irlb (double *A,                // Input data matrix (double case)
                   R = 1.0 / R_F;
                   R_F = 0;
                 }
-
               memmove (V + (j + 1) * n, F, n * sizeof (double));
               F77_NAME (dscal) (&n, &R, V + (j + 1) * n, &inc);
               B[j * work + j] = S;
@@ -387,7 +395,7 @@ irlb (double *A,                // Input data matrix (double case)
                   alpha = 1.0;
                   beta = 0.0;
                   F77_NAME (dgemv) ("n", &m, &n, &alpha, (double *) A, &m,
-                                    x, &inc, &beta, W + (j + 1) * m, &inc);
+                                    x, &inc, &beta, W + (j + 1) * m, &inc FCONE);
                 }
               mprod++;
               R_CheckUserInterrupt ();
@@ -414,7 +422,7 @@ irlb (double *A,                // Input data matrix (double case)
               orthog (W, W + (j + 1) * m, T, m, j + 1, 1);
               S = F77_NAME (dnrm2) (&m, W + (j + 1) * m, &inc);
               SS = 1.0 / S;
-              if (S < 2 * eps)
+              if (S < eps)
                 {
                   FOO = RNORM (m);
                   jj = (j + 1) * m;
@@ -439,12 +447,12 @@ irlb (double *A,                // Input data matrix (double case)
       memmove (BU, B, work * work * sizeof (double));   // Make a working copy of B
       int *BI = (int *) T;
       F77_NAME (dgesdd) ("O", &work, &work, BU, &work, BS, BU, &work, BV,
-                         &work, BW, &lwork, BI, &info);
+                         &work, BW, &lwork, BI, &info FCONE);
       R_F = F77_NAME (dnrm2) (&n, F, &inc);
       R = 1.0 / R_F;
       F77_NAME (dscal) (&n, &R, F, &inc);
 /* Force termination after encountering linear dependence */
-      if (R_F < 2 * eps)
+      if (R_F < eps)
         R_F = 0;
 
       Smax = 0;
@@ -458,6 +466,7 @@ irlb (double *A,                // Input data matrix (double case)
         res[kk] = R_F * BU[kk * work + (j - 1)];
 /* Update k to be the number of converged singular values. */
       convtests (j, nu, tol, svtol, Smax, svratio, res, &k, &converged, S);
+
       if (converged == 1)
         {
           iter++;
@@ -469,7 +478,7 @@ irlb (double *A,                // Input data matrix (double case)
       alpha = 1;
       beta = 0;
       F77_NAME (dgemm) ("n", "t", &n, &k, &j, &alpha, V, &n, BV, &work, &beta,
-                        V1, &n);
+                        V1, &n FCONE FCONE);
       memmove (V, V1, n * k * sizeof (double));
       memmove (V + n * k, F, n * sizeof (double));
 
@@ -484,7 +493,7 @@ irlb (double *A,                // Input data matrix (double case)
       alpha = 1;
       beta = 0;
       F77_NAME (dgemm) ("n", "n", &m, &k, &j, &alpha, W, &m, BU, &work, &beta,
-                        U1, &m);
+                        U1, &m FCONE FCONE);
       memmove (W, U1, m * k * sizeof (double));
       iter++;
     }
@@ -494,9 +503,9 @@ irlb (double *A,                // Input data matrix (double case)
   alpha = 1;
   beta = 0;
   F77_NAME (dgemm) ("n", "n", &m, &nu, &work, &alpha, W, &m, BU, &work, &beta,
-                    U, &m);
+                    U, &m FCONE FCONE);
   F77_NAME (dgemm) ("n", "t", &n, &nu, &work, &alpha, V, &n, BV, &work, &beta,
-                    V1, &n);
+                    V1, &n FCONE FCONE);
   memmove (V, V1, n * nu * sizeof (double));
 
   *ITER = iter;
@@ -528,6 +537,13 @@ __attribute__ ((visibility ("default")))
 void
 R_init_irlba (DllInfo * dll)
 {
+
+  R_RegisterCCallable("irlba", "orthog",
+                      (DL_FUNC) &orthog);
+  R_RegisterCCallable("irlba", "irlb",
+                      (DL_FUNC) &irlb);
+
+
   R_registerRoutines (dll, NULL, CallEntries, NULL, NULL);
   R_useDynamicSymbols (dll, 0);
   M_R_cholmod_start (&chol_c);
